/* Copyright 2008 BORDEAUX I UNIVERSITY & INRIA 
**
** This file is part of the PaStiX parallel sparse matrix package.
**
** This software is governed by the CeCILL-C license under French law
** and abiding by the rules of distribution of free software. You can
** use, modify and/or redistribute the software under the terms of the
** CeCILL-C license as circulated by CEA, CNRS and INRIA at the following
** URL: "http://www.cecill.info".
** 
** As a counterpart to the access to the source code and rights to copy,
** modify and redistribute granted by the license, users are provided
** only with a limited warranty and the software's author, the holder of
** the economic rights, and the successive licensors have only limited
** liability.
** 
** In this respect, the user's attention is drawn to the risks associated
** with loading, using, modifying and/or developing or reproducing the
** software by the user in light of its specific status of free software,
** that may mean that it is complicated to manipulate, and that also
** therefore means that it is reserved for developers and experienced
** professionals having in-depth computer knowledge. Users are therefore
** encouraged to load and test the software's suitability as regards
** their requirements in conditions enabling the security of their
** systems and/or data to be ensured and, more generally, to use and
** operate it in the same conditions as regards security.
** 
** The fact that you are presently reading this means that you have had
** knowledge of the CeCILL-C license and that you accept its terms.
*/
/****************************************
 * genheader.c
 *
 * generate an executable to generate
 * headers for pastix users.
 * Used by makefile
 ****************************************/

#include "common_pastix.h"
#include <errno.h>

#include <string.h>

/*
  Program: genheader

  Generate headers (for fortran and C ) setting INT and
  MPI_PASTIX_INT maccros to the correct value.

  Parameters:
  argc - must be 3
  argv - contains the names wanted for the C header and Fortran header

  Returns:
  EXIT_SUCCESS

*/
int main (int argc, char ** argv)
{
  char * header_c  = NULL;
  char * header_f  = NULL;
  char * header_m  = NULL;
  char * header_mp = NULL;
  char * murge_h   = NULL;
  char * genfort   = NULL;
  char * insertinc = NULL;
  char * cmd       = NULL;
  char * argument  = NULL;
  FILE * file      = NULL;
  int    cplx      = 0;
  int    realsize  = (int)(sizeof(FLOAT)/sizeof(byte));
  int    intsize   = (int)(sizeof(INT)/sizeof(byte));
#ifdef INTSSIZE64
  int    sintsize  = (int)(sizeof(int64_t)/sizeof(byte));
#else
  int    sintsize  = (int)(sizeof(int32_t)/sizeof(byte));
#endif
#ifdef TYPE_COMPLEX
  cplx     = 1;
  realsize = (int)(sizeof(FLOAT)/sizeof(byte)/2);
#endif

  if (argc != 9)
    {
      fprintf(stderr, "usage : %s headerNameC.h headerNameFortran.h"
              " headerMurge.inc headerMurgePastix.inc murge.h"
              " genfort.pl insert-fortran-inc.sh [C|Fortran|Murge]\n",argv[0]);
      return EXIT_FAILURE;
    }

  header_c  = argv[1];
  header_f  = argv[2];
  header_m  = argv[3];
  header_mp = argv[4];
  murge_h   = argv[5];
  genfort   = argv[6];
  insertinc = argv[7];
  argument  = argv[8];

  if (!strcmp(argument, "C")) {
    file = fopen (header_c,"w");
    if (file == NULL) {
      fprintf(stdout, "Error opening %s: %s\n", header_c, strerror(errno));
      abort();
    }

    /* #ifdef FORCE_COMPLEX */
    /* #if (defined X_ARCHalpha_compaq_osf1) */
    /* #ifndef   _RWSTD_HEADER_REQUIRES_HPP */
    /*   fprintf(file, "#include <complex>\n\n"); */
    /* #else  /\* _RWSTD_HEADER_REQUIRES_HPP *\/ */
    /*   fprintf(file, "#include <complex.hpp>\n\n"); */
    /* #endif /\* _RWSTD_HEADER_REQUIRES_HPP *\/ */
    /* #else /\*X_ARCHalpha_compaq_osf1*\/ */
    /*   fprintf(file, "#include <complex.h>\n\n"); */
    /* #endif */
    /* #endif */

#ifdef FORCE_LONG
    fprintf(file, "#define pastix_int_t     long\n");
    fprintf(file, "#define pastix_uint_t    unsigned long\n");
    fprintf(file, "#define MPI_PASTIX_INT   MPI_LONG\n");
#else
#ifdef FORCE_INT32
    fprintf(file, "#define pastix_int_t     int32_t\n");
    fprintf(file, "#define pastix_uint_t    u_int32_t\n");
    fprintf(file, "#define MPI_PASTIX_INT   MPI_INTEGER4\n");
#else
#ifdef FORCE_INT64
    fprintf(file, "#define pastix_int_t     int64_t\n");
    fprintf(file, "#define pastix_uint_t    u_int64_t\n");
    fprintf(file, "#define MPI_PASTIX_INT   MPI_INTEGER8\n");
#else
    fprintf(file, "#define pastix_int_t     int\n");
    fprintf(file, "#define pastix_uint_t    unsigned int\n");
    fprintf(file, "#define MPI_PASTIX_INT   MPI_INT\n");
#endif
#endif
#endif
#ifdef FORCE_DOUBLE
#ifdef FORCE_COMPLEX
    fprintf(file, "#if (defined _COMPLEX_H || defined _H_COMPLEX"
            " || defined __COMPLEX__)\n");
    fprintf(file,
      "#ifdef __cplusplus\n"
      "#ifndef COMPLEXDOUBLE_\n"
      "#define COMPLEXDOUBLE_\n"
      "struct{double realnum; double imagnum;} ComplexDouble_;\n"
      "#endif\n"
      "#define pastix_float_t ComplexDouble_\n"
      "#else\n"
      "#define pastix_float_t   double complex\n"
      "#endif\n");
    fprintf(file, "#define MPI_PASTIX_FLOAT GetMpiType()\n");
    fprintf(file, "#define FLOAT            pastix_float_t\n");
    fprintf(file, "#endif\n");
#else
    fprintf(file, "#define pastix_float_t   double\n");
    fprintf(file, "#define MPI_PASTIX_FLOAT MPI_DOUBLE\n");
    fprintf(file, "#define FLOAT            pastix_float_t\n");
#endif
#else
#ifdef FORCE_COMPLEX
    fprintf(file, "#if (defined _COMPLEX_H || defined _H_COMPLEX"
            " || defined __COMPLEX__ )\n");
    fprintf(file,
      "#ifdef __cplusplus\n"
      "#ifndef COMPLEXFLOAT_\n"
      "#define COMPLEXFLOAT_\n"
      "struct{float realnum; float imagnum;} ComplexFloat_;\n"
      "#endif\n"
      "#define pastix_float_t ComplexFloat_\n"
      "#else\n"
      "#define pastix_float_t float complex\n"
      "#endif\n");
    fprintf(file, "#define MPI_PASTIX_FLOAT GetMpiType()\n");
    fprintf(file, "#define FLOAT            pastix_float_t\n");
    fprintf(file, "#endif\n");
#else
    fprintf(file, "#define pastix_float_t   float\n");
    fprintf(file, "#define MPI_PASTIX_FLOAT MPI_FLOAT\n");
    fprintf(file, "#define FLOAT            pastix_float_t\n");
#endif
#endif
    fprintf(file, "#define INT              pastix_int_t\n");
    fprintf(file, "#define UINT             pastix_uint_t\n");
#ifdef FORCE_NOMPI
    fprintf(file, "#ifdef  MPI_Comm\n");
    fprintf(file, "#undef  MPI_Comm\n");
    fprintf(file, "#endif\n");
    fprintf(file, "#define MPI_Comm int\n");
#endif

    fclose(file);
  }
  else if (!strcmp(argument, "Fortran")) {

    /* Fortran header */
    file = fopen (header_f,"w");

    fprintf(file, "#define pastix_int_t       INTEGER(kind=%d)\n",
            (int)(sizeof(INT)/sizeof(byte)));
    fprintf(file, "#define pastix_uint_t      unsigned INTEGER(kind=%d)\n",
            (int)(sizeof(INT)/sizeof(byte)));
    fprintf(file, "#define pastix_data_ptr_t  INTEGER(kind=%d)\n",
            (int)(sizeof(INT*)/sizeof(byte)));
    fprintf(file, "#define MPI_PASTIX_INT     MPI_INTEGER%d\n",
            (int)(sizeof(INT)/sizeof(byte)));

#ifdef FORCE_COMPLEX
    fprintf(file, "#define pastix_float_t     COMPLEX(kind=%d)\n",
            (int)(sizeof(FLOAT)/sizeof(byte)/2));
    fprintf(file, "#define MPI_PASTIX_FLOAT   MPI_COMPLEX%d\n",
            (int)(sizeof(FLOAT)/sizeof(byte)));
#else
    fprintf(file, "#define pastix_float_t     REAL(kind=%d)\n",
            (int)(sizeof(FLOAT)/sizeof(byte)));
    fprintf(file, "#define MPI_PASTIX_FLOAT   MPI_REAL%d\n",
            (int)(sizeof(FLOAT)/sizeof(byte)));
#endif

    fclose(file);
  }
  else if (! strcmp(argument, "Murge")) {

    /* Murge Fortran header */
    if (NULL == (cmd = (char*)malloc((strlen(genfort)+
                                      strlen(murge_h)+
                                      strlen(header_m)+128)*sizeof(char))))
      return EXIT_FAILURE;
    sprintf(cmd, "perl ./%s -f %s -c %d -r %d -s %d -l %d > %s;",
            genfort, murge_h, cplx,
            realsize, sintsize, intsize, header_m);
    if (-1 == system(cmd))
      return EXIT_FAILURE;
    free(cmd);
    if (NULL == (cmd = (char*)malloc((strlen(header_mp)*3+
                                      strlen(insertinc)+
                                      strlen(header_m)+512)*sizeof(char))))
      return EXIT_FAILURE;
    sprintf(cmd, "sed -e 's/INTS/INTEGER(KIND=%d)/g'", sintsize);
#ifdef FORCE_COMPLEX
    sprintf(cmd, "%s  -e 's/COEF/COMPLEX(KIND=%d)/g'", cmd, realsize);
#else
    sprintf(cmd, "%s  -e 's/COEF/REAL(KIND=%d)/g'", cmd, realsize);
#endif
    sprintf(cmd, "%s %s > %s.tmp; %s %s %s.tmp;",cmd,
            header_mp, header_mp, insertinc, header_m, header_mp);
    fprintf(stdout, "%s\n", cmd);
    if (0 != system(cmd))
      return 1;
    free(cmd);
  }
  return 0;
}
